# -*- coding: utf-8 -*-

import os
import sys

from pathlib import Path
from PySide2.QtWidgets import *
from pytessng.MyPlugin import MyPlugin
from Tessng import *


def TessNgObject():
    app = QApplication()

    workspace = os.fspath(Path(__file__).resolve().parent)

    # 本地数据库
    config = {
        '__workspace': workspace,
        '__custsimubysteps': True,
        '__simuafterload': False,
        "__allowspopup": False,
    }

    plugin = MyPlugin()
    factory = TessngFactory()
    tessng = factory.build(plugin, config)

    if tessng is None:
        sys.exit(0)
    else:
        sys.exit(app.exec_())


if __name__ == '__main__':
    TessNgObject()



# -*- coding: utf-8 -*-
import csv
import json
import os
import traceback

from pathlib import Path
from DockWidget import *
from PySide2.QtWidgets import *
from Tessng import *
from threading import Thread
from xml.dom import minidom

from opendrive2tessng.utils.network_utils import Network
from tessng2other.opendrive.node import Doc
from tessng2other.opendrive.models import Junction, Connector, Road
from opendrive2tessng.main import main as TessNetwork
from pytessng.utils.functions import AdjustNetwork, line2surface


class MySignals(QObject):
    # 定义一种信号，因为有文本框和进度条两个类，此处要四个参数，类型分别是： QPlainTextEdit 、 QProgressBar、字符串和整形数字
    # 调用 emit方法发信号时，传入参数必须是这里指定的参数类型
    # 此处也可分开写两个函数，一个是文本框输出的，一个是给进度条赋值的
    text_print = Signal(QProgressBar, int, dict, bool)


class TESS_API_EXAMPLE(QMainWindow):
    def __init__(self, parent=None):
        super(TESS_API_EXAMPLE, self).__init__(parent)
        self.ui = Ui_TESS_API_EXAMPLEClass()
        self.ui.setupUi(self)
        self.createConnect()
        self.xodr = None
        self.network = None

    def createConnect(self):
        self.ui.btnOpenNet.clicked.connect(self.openNet)
        self.ui.btnCreateXodr.clicked.connect(self.createXodr)
        self.ui.btnCreateUnity.clicked.connect(self.createUnity)
        self.ui.btnShowXodr.clicked.connect(self.showXodr)
        self.ui.btnJoinLink.clicked.connect(self.joinLink)
        self.ui.btnSplitLink.clicked.connect(self.splitLink)
        self.ui.btnCreateLink.clicked.connect(self.createLink)
        self.ui.btnSimplifyTessngFile.clicked.connect(self.simplifyTessngFile)

    def createLink(self, info):
        iface = tngIFace()
        netiface = iface.netInterface()

        try:
            if self.ui.textCreateLink.text():
                point_1_x, point_1_y, point_2_x, point_2_y, *move = [float(i) for i in self.ui.textCreateLink.text().split(",")]
                move = move[0] if move else 0
                center_points = line2surface([(point_1_x, point_1_y, 0), (point_2_x, point_2_y, 0)], move)
                qt_center_points = [QPointF(m2p(point[0]), - m2p(point[1])) for point in center_points]

                netiface.createLink(qt_center_points, 1)
            else:
                message = "请参照提示信息输入\n起点横坐标,起点纵坐标,终点横坐标,终点纵坐标(,整体偏移距离(向右为正))"
                QMessageBox.warning(None, "提示信息", message)
        except:
            error = str(traceback.format_exc())
            print(error)
            QMessageBox.warning(None, "提示信息", '参数错误,请检查')

    def splitLink(self, info):
        iface = tngIFace()
        netiface = iface.netInterface()

        try:
            if self.ui.textSplitLink.text():
                split_infos = self.ui.textSplitLink.text().split(";")
                split_distances = []
                for split_info in split_infos:
                    link_id, point_x, point_y = split_info.split(",")
                    link_id, point_x, point_y = int(link_id), float(point_x), float(point_y)
                    locations = netiface.locateOnCrid(QPointF(m2p(point_x), -m2p(point_y)), 9)

                    for location in locations:
                        # 因为C++和python调用问题，必须先把lane实例化赋值给
                        if location.pLaneObject.isLane():
                            lane = location.pLaneObject.castToLane()
                            if lane.link().id() == link_id:
                                distance = location.distToStart
                                print("寻找到最近点", link_id, (point_x, point_y), location.point)
                                split_distances.append([link_id, distance])
                                break
                adjust_obj = AdjustNetwork(netiface)
                message = adjust_obj.split_link(split_distances)
                # if message and isinstance(message, str):
                #     QMessageBox.warning(None, "提示信息", message)
                self.ui.txtMessage1.setText(f"{message}")
            else:
                message = "请参照提示信息输入\n路段ID,断点横坐标,断点纵坐标;\n路段ID,断点横坐标,断点纵坐标...\n每组打断信息以< ; >分隔,内部以< , >分隔"
                QMessageBox.warning(None, "提示信息", message)
                return
        except:
            error = str(traceback.format_exc())
            print(error)
            QMessageBox.warning(None, "提示信息", '参数错误,请检查')

        # iface = tngIFace()
        # netiface = iface.netInterface()
        #
        # if not netiface.linkCount():
        #     return
        #
        # xodrSuffix = "OpenDrive Files (*.csv)"
        # dbDir = os.fspath(Path(__file__).resolve().parent / "Data")
        # file_path, filtr = QFileDialog.getOpenFileName(self, "打开文件", dbDir, xodrSuffix)
        # if not file_path:
        #     return
        # adjust_obj = AdjustNetwork(netiface)
        #
        # reader = csv.reader(open(file_path, 'r', encoding='utf-8'))
        # next(reader)
        # message = adjust_obj.split_link(reader)
        # if message and isinstance(message, str):
        #     QMessageBox.warning(None, "提示信息", message)
        # return

    def joinLink(self, info):
        iface = tngIFace()
        netiface = iface.netInterface()

        if not netiface.linkCount():
            return

        adjust_obj = AdjustNetwork(netiface)
        message = adjust_obj.join_link()
        # if message and isinstance(message, str):
        #     QMessageBox.warning(None, "提示信息", message)
        self.ui.txtMessage1.setText(f"{message}")
        return

    def createXodr(self, info):
        iface = tngIFace()
        netiface = iface.netInterface()

        if not netiface.linkCount():
            return

        xodrSuffix = "OpenDrive Files (*.xodr)"
        dbDir = os.fspath(Path(__file__).resolve().parent / "Data")
        file_path, filtr = QFileDialog.getSaveFileName(None, "文件保存", dbDir, xodrSuffix)
        if not file_path:
            return

        # 因为1.4 不支持多个前继/后续路段/车道，所以全部使用 junction 建立连接关系
        # 每个连接段视为一个 road，多个 road 组合成一个 junction
        connectors = []
        junctions = []
        for ConnectorArea in netiface.allConnectorArea():
            junction = Junction(ConnectorArea)
            junctions.append(junction)
            for connector in ConnectorArea.allConnector():
                # 为所有的 车道连接创建独立的road，关联至 junction
                for laneConnector in connector.laneConnectors():
                    connectors.append(Connector(laneConnector, junction))

        roads = []
        for link in netiface.links():
            roads.append(Road(link))

        # 路网绘制成功后，写入xodr文件
        doc = Doc()
        doc.init_doc()
        doc.add_junction(junctions)
        doc.add_road(roads + connectors)

        uglyxml = doc.doc.toxml()
        xml = minidom.parseString(uglyxml)
        xml_pretty_str = xml.toprettyxml()
        with open(file_path, 'w', encoding='utf-8') as f:
            f.write(xml_pretty_str)

        message = f"导出路网结构至 {file_path}"
        self.ui.txtMessage1.setText(message)

    def createUnity(self, info):
        iface = tngIFace()
        netiface = iface.netInterface()

        if not netiface.linkCount():
            return

        xodrSuffix = "OpenDrive Files (*.json)"
        dbDir = os.fspath(Path(__file__).resolve().parent / "Data")
        file_path, filtr = QFileDialog.getSaveFileName(None, "文件保存", dbDir, xodrSuffix)
        if not file_path:
            return

        # unity 信息提取
        from tessng2other.unity.unity_utils import convert_unity
        # TODO 车道类型相同，虚线，否则实线
        unity_info = convert_unity(netiface)
        unity_info = {'unity': unity_info, 'count': {}}
        for k, v in unity_info['unity'].items():
            unity_info['count'][k] = len(v)
        json.dump(unity_info, open(file_path, 'w'))

        message = f"导出路网结构至 {file_path}"
        self.ui.txtMessage1.setText(message)

    def openNet(self):
        xodrSuffix = "OpenDrive Files (*.xodr)"
        dbDir = os.fspath(Path(__file__).resolve().parent / "Data")

        iface = tngIFace()
        netiface = iface.netInterface()
        if not iface:
            return
        if iface.simuInterface().isRunning():
            QMessageBox.warning(None, "提示信息", "请先停止仿真，再打开路网")
            return

        count = netiface.linkCount()
        if count:
            # 关闭窗口时弹出确认消息
            reply = QMessageBox.question(self, '提示信息', '是否保存数据', QMessageBox.Yes, QMessageBox.No)
            # TODO 保存数据--> 清除数据 --> 打开新文件
            if reply == QMessageBox.Yes:
                netiface.saveRoadNet()

        # custSuffix = "TESSNG Files (*.tess);;TESSNG Files (*.backup);;OpenDrive Files (*.xodr)"
        netFilePath, filtr = QFileDialog.getOpenFileName(self, "打开文件", dbDir, xodrSuffix)
        print(f"导入的 opendrive 路网: {netFilePath}")
        if not netFilePath:
            return
        self.xodr = netFilePath
        # 限制文件的再次选择
        self.ui.btnOpenNet.setEnabled(False)
        # 声明线程间的共享变量
        global pb
        global my_signal
        my_signal = MySignals()
        pb = self.ui.pb

        step_length = float(self.ui.xodrStep.currentText().split(" ")[0])
        self.network = TessNetwork(netFilePath)

        # 主线程连接信号
        my_signal.text_print.connect(self.ui.change_progress)
        # 启动子线程
        context = {
            "signal": my_signal.text_print,
            "pb": pb
        }
        filters = None  # list(LANE_TYPE_MAPPING.keys())
        thread = Thread(target=self.network.convert_network, args=(step_length, filters, context))
        thread.start()

    def showXodr(self, info):
        """
        点击按钮，绘制 opendrive 路网
        Args:
            info: None
        Returns:
        """
        if not (self.network and self.network.network_info):
            QMessageBox.warning(None, "提示信息", "请先导入xodr路网文件或等待文件转换完成")
            return

        # 代表TESS NG的接口
        tess_lane_types = []
        for xodrCk in self.ui.xodrCks:
            if xodrCk.checkState() == QtCore.Qt.CheckState.Checked:
                tess_lane_types.append(xodrCk.text())
        if not tess_lane_types:
            QMessageBox.warning(None, "提示信息", "请至少选择一种车道类型")
            return

        # # 简单绘制路网走向
        # from matplotlib import pyplot as plt
        # for value in self.network.network_info['roads_info'].values():
        #     for points in value['road_points'].values():
        #         x = [i['position'][0] for i in points['right_points']]
        #         # x = [point['right_points'][['position']][0] for point in points]
        #         y = [i['position'][1] for i in points['right_points']]
        #         plt.plot(x, y)
        # plt.show()

        # 打开新底图
        iface = tngIFace()
        netiface = iface.netInterface()
        attrs = netiface.netAttrs()
        if attrs is None or attrs.netName() != "PYTHON 路网":
            netiface.setNetAttrs("PYTHON 路网", "OPENDRIVE", otherAttrsJson=self.network.network_info["header_info"])

        error_junction = self.network.create_network(tess_lane_types, netiface)
        message = "\n".join([str(i) for i in error_junction])

        self.ui.txtMessage2.setText(f"{message}")
        is_show = bool(error_junction)
        self.ui.text_label_2.setVisible(is_show)
        self.ui.txtMessage2.setVisible(is_show)

    # 简化路网
    def simplifyTessngFile(self, info):
        iface = tessngIFace()
        netiface = iface.netInterface()
        netFilePath = netiface.netFilePath()

        if not netFilePath.endswith(".tess"):
            QMessageBox.warning(None, "提示信息", "请打开合适的 tess 路网")
            return

        try:
            message = Network.simplify_tessng_file(netFilePath)
            message += "\n\n注意: 原文件已被简化, 请重新打开"
            self.ui.txtMessage1.setText(message)
        except:
            QMessageBox.warning(None, "提示信息", "路网简化失败, 请联系开发者")
            return


if __name__ == '__main__':
    app = QApplication()
    win = TESS_API_EXAMPLE()
    win.show()
    app.exec_()



from pytessng import Tessng
from Tessng import *


# 用户插件子类，代表用户自定义与路网相关的实现逻辑，继承自MyCustomerNet
class MyNet(PyCustomerNet):
    def __init__(self):
        super(MyNet, self).__init__()

    # 创建路网
    def createNet(self):
        pass

    def ref_curvatureMinDist(self, itemType: int, itemId: int, ref_minDist: Tessng.objreal):
        ref_minDist.value = 0.1
        return True

    def afterLoadNet(self):
        # 代表TESS NG的接口
        iface = tessngIFace()
        # 代表TESS NG的路网子接口
        netiface = iface.netInterface()
        # 初始网格化
        netiface.buildNetGrid(30)


import collections
import csv
from math import sqrt

import numpy as np

from PySide2.QtGui import *
from Tessng import *
from numpy import square


def qtpoint2point(qtpoints):
    points = []
    for qtpoint in qtpoints:
        points.append(
            [m2p(qtpoint.x()), - m2p(qtpoint.y()), m2p(qtpoint.z())] if isinstance(qtpoint, QVector3D) else qtpoint
        )
    return points


def point2qtpoint(points):
    qtpoints = []
    for point in points:
        qtpoints.append(
            QVector3D(p2m(point[0]), - p2m(point[1]), p2m(point[2])) if not isinstance(point, QVector3D) else point
        )
    return qtpoints

def line2surface(base_points, move_parameter):
    """
        根据基础点序列，向右偏移一定距离)
    """
    points = []
    point_count = len(base_points)
    for index in range(point_count):
        if index + 1 == point_count:
            is_last = True
            num = index - 1
        else:
            is_last = False
            num = index

        point = deviation_point(base_points[num], base_points[num + 1], move_parameter, is_last=is_last)
        points.append([point[0], point[1], base_points[index][2]])
    return points

def deviation_point(coo1, coo2, move_parameter, is_last=False):
    x1, y1, x2, y2 = coo1[0], coo1[1], coo2[0], coo2[1]  # 如果是最后一个点，取第二个 点做偏移
    x_base, y_base = (x1, y1) if not is_last else (x2, y2)
    if not ((x2 - x1) or (y2 - y1)):  # 分母为0
        return [x_base, y_base, 0]
    X = x_base + move_parameter * (y2 - y1) / sqrt(square(x2 - x1) + square((y2 - y1)))
    Y = y_base + move_parameter * (x1 - x2) / sqrt(square(x2 - x1) + square((y2 - y1)))
    return [X, Y, 0]


class Road:
    def __init__(self):
        self.link = None
        self.last_links = []
        self.next_links = []
        self.connectors = []  # 暂时没用到


class AdjustNetwork:
    def __init__(self, netiface):
        self.netiface = netiface
        self.roads = self.calc_connector()

        self.connector_area_mapping = self.calc_connector_area()

        # 在调整的过程中，旧的link在不断删除，同时新的link被创建，所以需要建立映射关系表
        # connector也有可能在消失，所以不再被记录，通过上下游link获取connector
        self.old_new_link_mapping = {}
        # 初始化映射表,link在调整前后可能存在一对多/多对一关系，所以用 dict&list 记录
        for link in netiface.links():
            self.old_new_link_mapping[link.id()] = [link.id()]

    # 记录全域的连接段面域
    def calc_connector_area(self):
        connector_area_mapping = collections.defaultdict(list)
        for ConnectorArea in self.netiface.allConnectorArea():
            for connector in ConnectorArea.allConnector():
                connector_area_mapping[ConnectorArea.id()].append(connector.id())
        return connector_area_mapping


    # 记录全局的连接关系
    def calc_connector(self):
        roads = {}
        for connector in self.netiface.connectors():
            # 上下游不可能为空
            last_link = connector.fromLink()
            next_link = connector.toLink()

            last_road = roads.get(last_link.id(), Road())
            next_road = roads.get(next_link.id(), Road())

            last_road.link = last_link
            last_road.next_links.append(next_link.id())
            last_road.connectors.append(connector)

            next_road.link = next_link
            next_road.last_links.append(last_link.id())
            next_road.connectors.append(connector)

            roads[next_link.id()] = next_road
            roads[last_link.id()] = last_road

        # 记录无连接段的路段
        for link in self.netiface.links():
            if link.id() not in roads:
                road = Road()
                road.link = link
                roads[link.id()] = road
        return roads

    # 计算路段上所有的路段点，车道点
    # 根据已知的信息计算新的断点序列列表
    @staticmethod
    def calc_points(points, indexs_list, ratios):
        last_index = 0
        points = qtpoint2point(points)

        new_points_list = []  # [[] for _ in indexs_list]

        first_point = None
        # 先把初识点全部分配下去
        for _, indexs in enumerate(indexs_list):
            if indexs:
                last_index = indexs[-1]  # 更新计算点
            new_points = [points[index] for index in indexs]
            # 添加首尾点
            if first_point is not None:
                new_points.insert(0, first_point)

            ratio = ratios[_]
            final_point = np.array(points[last_index]) * (1 - ratio) + np.array(points[last_index + 1]) * ratio
            new_points.append(final_point)
            first_point = final_point

            new_points_list.append(new_points)
            # if indexs:
            #     last_index = indexs[-1]  # 更新计算点

        # 添加最后一个路段，第一个点为上一路段计算的终点，第二个点为上一路段所在点序列的下一个点，最后一个点为路段终点
        new_points_list.append([first_point] + points[last_index + 1:])
        new_points_list = [point2qtpoint(i) for i in new_points_list]
        return new_points_list

    def calc_split_parameter(self, link, split_link_info):
        center_points = link.centerBreakPoint3Ds()
        center_points = qtpoint2point(center_points)

        sum_length = 0
        last_x, last_y, last_z = center_points[0]
        lengths = [0] + split_link_info["lengths"]
        for point_index, point in enumerate(center_points):
            x, y, z = point
            distance = np.sqrt((x - last_x) ** 2 + (y - last_y) ** 2)
            last_x, last_y, last_z = x, y, x

            new_sum_length = sum_length + distance
            for split_index, split_length in enumerate(lengths):
                if split_index == 0:
                    continue
                if new_sum_length < lengths[split_index - 1]:
                    continue
                elif new_sum_length >= lengths[split_index - 1] and new_sum_length < lengths[split_index]:
                    # 区间命中
                    split_link_info['index'][split_index - 1].append(point_index)
                    # TODO 区间已经命中，可以 break 了
                # elif sum_length >= lengths[split_index - 1] and new_sum_length > split_length:
                #     # 新点超出范围,计算比例
                #     ratio = (split_length - sum_length) / distance
                #     split_link_info['ratio'][split_index - 1] = ratio  # 此处不需要记录 首点的比例，因为当需要用到首点时，采用上一点的断点计算
                else:  # new_sum_length >= lengths[split_index]
                    # 如果此处并没有 ratio，说明这是第一次被匹配上，需要进行ratio 计算
                    if split_link_info['ratio'][split_index - 1] is None:
                        ratio = (split_length - sum_length) / distance
                        split_link_info['ratio'][split_index - 1] = ratio  # 此处不需要记录 首点的比例，因为当需要用到首点时，采用上一点的断点计算
            sum_length = new_sum_length

        # 计算完成，可以进行分割
        if len(split_link_info['lengths']) != len(split_link_info['index']) or len(split_link_info['lengths']) != len(
                split_link_info['index']):
            raise 1  # 初步判断

    def calc_split_links_info(self, link, split_link_info):
        # 计算新的 link 信息
        indexs_list = split_link_info['index']
        ratios = split_link_info['ratio']
        link_center_points = self.calc_points(link.centerBreakPoint3Ds(), indexs_list, ratios)

        new_links_info = [
            {
                'center': link_center_points[index],
                # 'name': f"{link.name()}/{index}",
                'name': f"{link.name()}",  # 保留原路段名
                'lanes': collections.defaultdict(lambda: {
                    'center': [],
                    'left': [],
                    'right': [],
                    'type': '',
                    'attr': {},
                }),
                'old_link_id': link.id(),
            } for index in range(len(indexs_list) + 1)
        ]

        for lane in link.lanes():
            center_points = self.calc_points(lane.centerBreakPoint3Ds(), indexs_list, ratios)
            left_points = self.calc_points(lane.leftBreakPoint3Ds(), indexs_list, ratios)
            right_points = self.calc_points(lane.rightBreakPoint3Ds(), indexs_list, ratios)
            for index in range(len(indexs_list) + 1):  # 被分割后的 link 数量,比分割点数大 1
                new_links_info[index]['lanes'][lane.number()] = {
                    'center': center_points[index],
                    'left': left_points[index],
                    'right': right_points[index],
                    'type': lane.actionType(),
                    'attr': {},
                }
        return new_links_info

    def split_link(self, reader):
        split_links_info = collections.defaultdict(lambda: {'lengths': [], 'index': [], 'ratio': []})
        for row in reader:
            try:
                row = [float(i) for index, i in enumerate(row)]
            except:
                return f"输入数据错误:{row}"

            link_id = int(row[0])
            points_length = sorted(row[1:])
            link = self.netiface.findLink(link_id)
            if not link:
                return f"link: {link_id} 不存在"
            if min(points_length) <= 0 or max(points_length) >= link.length() or len(points_length) != len(
                    set(points_length)):
                return f"link: {row[0]} 长 {link.length()}, 断点长度输入不准确"

            split_links_info[link_id]['lengths'] = points_length
            split_links_info[link_id]['index'] = [[] for _ in points_length]
            split_links_info[link_id]['ratio'] = [None for _ in points_length]

        # 路段有可能发生 合并，切分，所以使用 list 存储最为合理
        old_new_link_mapping = collections.defaultdict(list)
        for link in self.netiface.links():
            old_new_link_mapping[link.id()].append(link.id())

        old_connectors = []
        # 记录原始连接信息
        for link_id in split_links_info.keys():
            # last_link, next_link 都不属于 link_group,所以在过程中可能已经被删除，只能通过id映射获取新的link
            # 可能存在多个上游路段和多个下游路段，需要逐个获取连接段，从而记录连接关系
            old_connectors += self.calc_upstream_connection(link_id)
            old_connectors += self.calc_downstream_connection(link_id)

        all_new_link_info = []
        # 计算新的路段信息
        for link_id, split_link_info in split_links_info.items():
            link = self.netiface.findLink(link_id)
            # 根据 link 实际信息 丰富 切割参数 > split_link_info
            self.calc_split_parameter(link, split_link_info)
            # 获取切割详情，含点序列等基本信息
            new_links_info = self.calc_split_links_info(link, split_link_info)
            all_new_link_info += new_links_info
            # 记录 link 基本信息后移除
            old_link_id = link.id()
            old_new_link_mapping[old_link_id].remove(old_link_id)  # 删除路段前，移除相关的映射关系
            self.netiface.removeLink(link)

        temp_link_mapping = collections.defaultdict(list)
        # 根据记录的信息, 集中创建新的路段并更新映射表
        for new_link_info in all_new_link_info:
            # 做路段分割时，原始路段有且仅有一个
            old_link_id = new_link_info['old_link_id']
            new_link_obj = self.create_new_link(new_link_info)
            # 记录进映射表
            old_new_link_mapping[old_link_id].append(new_link_obj.id())
            temp_link_mapping[old_link_id].append(new_link_obj.id())

        message = "路段分割结果: \n"
        for k, v in temp_link_mapping.items():
            message += f"{k} --> {v} \n"

        # 集中创建内部的连接段
        for old_link_id, new_link_ids in old_new_link_mapping.items():
            for index in range(len(new_link_ids) - 1):
                from_link_id = old_new_link_mapping[old_link_id][index]
                to_link_id = old_new_link_mapping[old_link_id][index + 1]
                link_obj = self.netiface.findLink(from_link_id)
                # 被分割的小路段的车道数/车道类型均完全一致，所以随便取一个路段即可
                lanes = [lane.number() + 1 for lane in link_obj.lanes()]
                self.netiface.createConnector(from_link_id, to_link_id, lanes, lanes, f"{from_link_id}-{to_link_id}")

        # 根据记录的信息 批量创建新的连接段
        self.create_all_new_connector(old_connectors, old_new_link_mapping)
        return message

    def join_link(self):
        # TODO 合并路段, 需要添加车道类型判断，后续增加点位自合并的方法
        link_groups = []
        exist_links = []
        for link_id, road in self.roads.items():
            if link_id in exist_links:
                # 已经进行过查找的 link 不需要再次遍历
                continue

            # 获取 link 相应的 上下游link 并组成有序列表
            link_group = [road.link]
            self.get_chain_by_next(road, link_group)
            self.get_chain_by_last(road, link_group)

            link_groups.append(link_group)
            exist_links += [i.id() for i in link_group]

        # 判断是否有路段进行过重复查询，如果有，说明逻辑存在漏洞
        if len(exist_links) != len(set(exist_links)):
            return "出现唯一性错误，请联系开发者"

        # 在调整的过程中，旧的link在不断删除，同时新的link被创建，所以需要建立映射关系表，connector也有可能在消失，所以不再被记录，通过上下游link获取connector
        # 初始化映射表, 为了和 分割路段保持一致，仍然采用 dict(list) 的形式，但实际上, 最后的 list 有且仅有一个值
        old_new_link_mapping = collections.defaultdict(list)
        for link in self.netiface.links():
            old_new_link_mapping[link.id()].append(link.id())

        # TODO 根据信息做路网调整, 在调整过程中，未遍历到的 link_group 不会被调整，即不会丢失对象
        # 分步做，先统计原始的连接段信息，方便后续迭代
        old_connectors = []
        for link_group in filter(lambda x: len(x) > 1, link_groups):
            # 记录原始信息，方便后续重新创建路段及连接段
            first_link = link_group[0]
            final_link = link_group[-1]

            # last_link, next_link 都不属于 link_group,所以在过程中可能已经被删除，只能通过id映射获取新的link
            # 可能存在多个上游路段和多个下游路段，需要逐个获取连接段，从而记录连接关系
            old_connectors += self.calc_upstream_connection(first_link.id())
            old_connectors += self.calc_downstream_connection(final_link.id())

        all_new_link_info = []
        # 记录路段合并信息
        for link_group in filter(lambda x: len(x) > 1, link_groups):
            new_link_info = {
                'center': [],
                'name': '',
                'lanes': collections.defaultdict(lambda: {
                    'center': [],
                    'left': [],
                    'right': [],
                    'type': '',
                    'attr': {},
                }),
                'old_link_ids': [i.id() for i in link_group],
            }

            # 先记录id
            for link in link_group:  # 有序的进行点位合并
                # TODO 暂时不记录中间连接段的点序列
                new_link_info['center'] += link.centerBreakPoint3Ds()
                new_link_info['name'] += link.name()
                for lane in link.lanes():
                    lane_number = lane.number()
                    new_link_info['lanes'][lane_number]['center'] += lane.centerBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['left'] += lane.leftBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['right'] += lane.rightBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['type'] = lane.actionType()

                # 记录 link 基本信息后移除
                old_new_link_mapping[link.id()].remove(link.id())
                self.netiface.removeLink(link)
            all_new_link_info.append(new_link_info)

        message = "路段合并结果: \n"
        # 集中创建新的路段并更新映射表
        for new_link_info in all_new_link_info:
            new_link_obj = self.create_new_link(new_link_info)
            message += f"{new_link_info['old_link_ids']} --> {new_link_obj.id()} \n"
            # 更新映射表，原本多个旧路段都指向了同一个新路段
            for old_link_id in new_link_info['old_link_ids']:
                old_new_link_mapping[old_link_id].append(new_link_obj.id())

        # 创建新的连接段,
        # 如果某连接段上下游均进行了路段合并，则连接段会被重新重复创建，已被过滤
        self.create_all_new_connector(old_connectors, old_new_link_mapping)

        return message

    # 计算路段的上游连接关系
    def calc_upstream_connection(self, link_id):
        road = self.roads[link_id]
        connectors = []
        for last_link_id in road.last_links:
            connector_info = self.get_connector_info(last_link_id, link_id)
            connectors.append(connector_info)
        return connectors

    # 计算路段的下游连接关系
    def calc_downstream_connection(self, link_id):
        road = self.roads[link_id]
        connectors = []
        for next_link_id in road.next_links:
            connector_info = self.get_connector_info(link_id, next_link_id)
            connectors.append(connector_info)
        return connectors

    def get_connector_info(self, from_link_id, to_link_id):
        connector = self.netiface.findConnectorByLinkIds(from_link_id, to_link_id)
        return {
            'from_link_id': from_link_id,
            'to_link_id': to_link_id,
            'connector': [
                (i.fromLane().number(), i.toLane().number())
                for i in connector.laneConnectors()
            ],
            'lanesWithPoints3': [
                {
                    "center": i.centerBreakPoint3Ds(),
                    "left": i.leftBreakPoint3Ds(),
                    "right": i.rightBreakPoint3Ds(),
                }
                for i in connector.laneConnectors()
            ],
        }

    def get_chain_by_next(self, road, link_group: list):
        if len(road.next_links) != 1:
            # 有且仅有一个下游，才可以继续延伸
            return

        next_link_id = road.next_links[0]
        # 新增判断，即使路段只有一个下游，连接段所属面域中存在多个连接段，仍然不允许合并
        connector = self.netiface.findConnectorByLinkIds(road.link.id(), next_link_id)
        for value in self.connector_area_mapping.values():
            if connector.id() in value and len(value) > 1:
                print(f"匹配逻辑出现唯一性错误 {value}")
                return

        next_link = self.netiface.findLink(next_link_id)
        next_road = self.roads[next_link.id()]
        # 判断下游 link 是否有且仅有 1 个上游，且车道数/车道类型与当前link一致，若一致，加入链路并继续向下游寻找
        if len(next_road.last_links) == 1 and [lane.actionType() for lane in road.link.lanes()] == [
            lane.actionType() for lane in next_road.link.lanes()]:
            link_group.append(next_link)
            self.get_chain_by_next(next_road, link_group)
        return

    # 通过指定路段信息寻找匹配的上下游
    def get_chain_by_last(self, road, link_group: list):
        if len(road.last_links) != 1:
            # 有且仅有一个上游，才可以继续延伸
            return
        last_link_id = road.last_links[0]
        # 新增判断，即使路段只有一个下游，连接段所属面域中存在多个连接段，仍然不允许合并
        connector = self.netiface.findConnectorByLinkIds(last_link_id, road.link.id())
        for value in self.connector_area_mapping.values():
            if connector.id() in value and len(value) > 1:
                print(f"匹配逻辑出现唯一性错误 {value}")
                return

        last_link = self.netiface.findLink(last_link_id)
        last_road = self.roads[last_link.id()]
        # 判断上游 link 是否有且仅有 1 个下游，且车道数与当前link一致，若一致，加入链路并继续向上游寻找
        if len(last_road.next_links) == 1 and [lane.actionType() for lane in road.link.lanes()] == [
            lane.actionType() for lane in last_road.link.lanes()]:
            link_group.insert(0, last_link)
            self.get_chain_by_last(last_road, link_group)
        return

    def create_new_link(self, new_link_info):
        new_link_obj = self.netiface.createLink3DWithLanePointsAndAttrs(
            new_link_info['center'],
            [
                {
                    'center': new_link_info['lanes'][k]['center'],
                    'right': new_link_info['lanes'][k]['right'],
                    'left': new_link_info['lanes'][k]['left'],
                } for k in sorted(new_link_info['lanes'])
            ],  # 必须排序
            [new_link_info['lanes'][k]['type'] for k in sorted(new_link_info['lanes'])],
            [new_link_info['lanes'][k]['attr'] for k in sorted(new_link_info['lanes'])],
            new_link_info['name']
        )
        return new_link_obj

    # 根据记录的信息 批量创建新的连接段
    def create_all_new_connector(self, old_connectors, old_new_link_mapping):
        # 如果某连接段上下游均进行了路段合并，则连接段会被重新重复创建，此处进行过滤
        exist_connector = []
        for connector in old_connectors:
            old_from_link_id = connector['from_link_id']
            old_to_link_id = connector['to_link_id']
            new_from_link_id = old_new_link_mapping[old_from_link_id][-1]  # 上游路段取新的link列表的最后一个
            new_to_link_id = old_new_link_mapping[old_to_link_id][0]  # 下游路段取新的link列表的第一个

            connector_name = f'{new_from_link_id}_{new_to_link_id}'
            if connector_name in exist_connector:
                continue
            self.netiface.createConnector3DWithPoints(new_from_link_id,
                                                      new_to_link_id,
                                                      [i[0] + 1 for i in connector['connector']],
                                                      [i[1] + 1 for i in connector['connector']],
                                                      connector['lanesWithPoints3'],
                                                      f"{new_from_link_id}-{new_to_link_id}"
                                                      )
            # 采用默认连接，防止出现连接段不平滑问题
            # netiface.createConnector(new_from_link_id, new_to_link_id, [i[0] + 1 for i in connector['connector']], [i[1] + 1 for i in connector['connector']], f"{new_from_link_id}-{new_to_link_id}")
            exist_connector.append(connector_name)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'TESS_API_EXAMPLE.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################
from PySide2 import QtCore
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from opendrive2tessng.utils.config import WIDTH_LIMIT, LANE_TYPE_MAPPING


class Ui_TESS_API_EXAMPLEClass(object):
    def setupUi(self, TESS_API_EXAMPLEClass):
        if not TESS_API_EXAMPLEClass.objectName():
            TESS_API_EXAMPLEClass.setObjectName(u"TESS_API_EXAMPLEClass")
        TESS_API_EXAMPLEClass.resize(262, 735)

        self.centralWidget = QWidget(TESS_API_EXAMPLEClass)
        self.centralWidget.setObjectName(u"centralWidget")
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")

        # 文件选择框
        self.groupBox_1 = QGroupBox(self.centralWidget)
        self.groupBox_1.setObjectName(u"groupBox_1")
        self.verticalLayout_0 = QVBoxLayout(self.groupBox_1)

        # 其他选项
        self.groupBox_4 = QGroupBox(self.centralWidget)
        self.groupBox_4.setObjectName(u"groupBox_out")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox_4)

        xodr_label1 = QLabel()
        xodr_label1.setText("路段最小分段长度(请在文件导入前选择)")
        self.xodrStep = QComboBox(self.centralWidget)
        self.xodrStep.addItems(("1.0 m", "0.5 m", "5.0 m", "10 m", "20 m"))
        # self.xodrStep.addItems(("5 m", "50 m", "100 m", "1000 m"))

        # 文件导入进度条
        self.pb = QProgressBar(self.centralWidget)
        self.pb.setRange(0, 100)  # 进度对话框的范围设定
        self.pb.setTextVisible(False)

        self.btnOpenNet = QPushButton(self.centralWidget)
        self.btnOpenNet.setObjectName(u"btnOpenNet")

        # 导出 opendrive 文件
        self.btnCreateXodr = QPushButton(self.centralWidget)
        self.btnCreateXodr.setObjectName(u"btnCreateXodr")

        # 导出 unity 文件
        self.btnCreateUnity = QPushButton(self.centralWidget)
        self.btnCreateUnity.setObjectName(u"btnCreateUnity")

        # 路网文件自调整
        self.btnJoinLink = QPushButton(self.centralWidget)
        self.btnJoinLink.setObjectName(u"btnJoinLink")

        # 路段断开
        # 允许用户输入路段ID以及点位坐标，如若为空，拉起文件选择框
        self.textSplitLink = QLineEdit(self.centralWidget)
        self.textSplitLink.setPlaceholderText("link_id,x,y;link_id,x,y...")
        # self.textSplitLink.setObjectName(u"textSplitLink")
        self.btnSplitLink = QPushButton(self.centralWidget)
        self.btnSplitLink.setObjectName(u"btnSplitLink")

        # 新建路段
        self.textCreateLink = QLineEdit(self.centralWidget)
        self.textCreateLink.setPlaceholderText("x1,y1,x2,y2(,move)")
        # self.textSplitLink.setObjectName(u"textSplitLink")
        self.btnCreateLink = QPushButton(self.centralWidget)
        self.btnCreateLink.setObjectName(u"btnCreateLink")

        # 简化路网
        self.btnSimplifyTessngFile = QPushButton(self.centralWidget)
        self.btnSimplifyTessngFile.setObjectName(u"btnSimplifyTessngFile")

        # self.verticalLayout_0.addWidget(self.pd)
        self.verticalLayout_0.addWidget(xodr_label1)
        self.verticalLayout_0.addWidget(self.xodrStep)
        self.verticalLayout_0.addWidget(self.pb)
        self.verticalLayout_0.addWidget(self.btnOpenNet)
        self.verticalLayout_4.addWidget(self.btnCreateXodr)
        self.verticalLayout_4.addWidget(self.btnCreateUnity)
        self.verticalLayout_4.addWidget(self.btnJoinLink)
        self.verticalLayout_4.addWidget(self.textSplitLink)
        self.verticalLayout_4.addWidget(self.btnSplitLink)
        self.verticalLayout_4.addWidget(self.textCreateLink)
        self.verticalLayout_4.addWidget(self.btnCreateLink)
        self.verticalLayout_4.addWidget(self.btnSimplifyTessngFile)
        self.groupBox_4.setVisible(True)

        # 信息窗
        self.groupBox_3 = QGroupBox(self.centralWidget)
        self.groupBox_3.setObjectName(u"groupBox_3")

        self.verticalLayout_2 = QVBoxLayout(self.groupBox_3)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(1, -1, 1, -1)

        self.text_label_1 = QLabel()
        self.text_label_1.setText("路网详情")
        self.txtMessage1 = QTextBrowser(self.groupBox_3)
        self.txtMessage1.setObjectName(u"txtMessage")

        self.text_label_2 = QLabel()
        self.text_label_2.setText("创建异常信息提示窗\n(用户可根据异常信息手动更改)")
        self.txtMessage2 = QTextBrowser(self.groupBox_3)
        self.txtMessage2.setObjectName(u"txtMessage")

        self.verticalLayout_2.addWidget(self.text_label_1)
        self.verticalLayout_2.addWidget(self.txtMessage1)
        self.verticalLayout_2.addWidget(self.text_label_2)
        self.verticalLayout_2.addWidget(self.txtMessage2)

        # xodr 创建选择页
        self.groupBox_2 = QGroupBox(self.centralWidget)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox_2)

        xodr_label2 = QLabel()
        xodr_label2.setText("导入车道类型选择")

        self.xodrCks = [QCheckBox(LANE_TYPE) for LANE_TYPE in set(LANE_TYPE_MAPPING.values())]
        for i in self.xodrCks:
            i.setCheckState(QtCore.Qt.Checked)

        self.btnShowXodr = QPushButton(self.centralWidget)
        self.btnShowXodr.setObjectName(u"btnShowXodr")

        xodr_label3 = QLabel()

        context = "车道转换说明:\n"
        for lane_type, limit_data in WIDTH_LIMIT.items():
            split_limit = limit_data.get('split', None)
            join_limit = limit_data.get('join', None)
            context += f"{lane_type}:\n{f'小于 {join_limit}m'.ljust(12)}:不解析\n{f'{join_limit}m 至 {split_limit}m'.ljust(13)}:视为连接段\n{f'大于 {split_limit}m'.ljust(12)}:正常车道\n"
        # 中文字符宽度是英文字符的两倍
        xodr_label3.setText(context)

        self.verticalLayout_4.addWidget(xodr_label2)
        for xodrCk in self.xodrCks:
            self.verticalLayout_4.addWidget(xodrCk)
        # self.verticalLayout_4.addWidget(self.xodrCk2)
        self.verticalLayout_4.addWidget(self.btnShowXodr)
        self.verticalLayout_4.addWidget(xodr_label3)

        # 添加控件到布局
        self.verticalLayout.addWidget(self.groupBox_1)
        self.verticalLayout.addWidget(self.groupBox_2)
        self.verticalLayout.addWidget(self.groupBox_3)
        self.verticalLayout.addWidget(self.groupBox_4)

        self.groupBox_3.setVisible(True)  # 信息窗
        self.pb.setVisible(False)  # 信息窗
        self.groupBox_2.setVisible(False)  # 创建选择框
        self.text_label_1.setVisible(False)
        self.text_label_2.setVisible(False)
        self.txtMessage1.setVisible(True)  # error 信息栏
        self.txtMessage2.setVisible(False)  # error 信息栏
        # xodr 控件结束

        TESS_API_EXAMPLEClass.setCentralWidget(self.centralWidget)
        self.menuBar = QMenuBar(TESS_API_EXAMPLEClass)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 262, 26))
        TESS_API_EXAMPLEClass.setMenuBar(self.menuBar)
        self.mainToolBar = QToolBar(TESS_API_EXAMPLEClass)
        self.mainToolBar.setObjectName(u"mainToolBar")
        TESS_API_EXAMPLEClass.addToolBar(Qt.TopToolBarArea, self.mainToolBar)
        self.statusBar = QStatusBar(TESS_API_EXAMPLEClass)
        self.statusBar.setObjectName(u"statusBar")
        TESS_API_EXAMPLEClass.setStatusBar(self.statusBar)

        self.retranslateUi(TESS_API_EXAMPLEClass)

        QMetaObject.connectSlotsByName(TESS_API_EXAMPLEClass)

    def retranslateUi(self, TESS_API_EXAMPLEClass):
        TESS_API_EXAMPLEClass.setWindowTitle(
            QCoreApplication.translate("TESS_API_EXAMPLEClass", u"TESS_API_EXAMPLE", None))
        self.btnOpenNet.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u9009\u62e9\u6587\u4ef6", None))
        self.btnCreateXodr.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"导出opendrive文件", None))
        self.btnCreateUnity.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"导出unity文件", None))
        self.btnJoinLink.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"路段合并\n主动合并路网内的全部路段", None))
        self.btnSplitLink.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"路段断开", None))
        self.btnCreateLink.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"路段创建", None))
        self.btnSimplifyTessngFile.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"简化Tess路网", None))
        # self.btnStartSimu.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u542f\u52a8\u4eff\u771f", None))
        # self.btnPauseSimu.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u6682\u505c\u4eff\u771f", None))
        # self.btnStopSimu.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u505c\u6b62\u4eff\u771f", None))
        self.groupBox_1.setTitle(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"opendrive文件导入", None))
        self.groupBox_4.setTitle(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"其他功能", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u4fe1\u606f\u7a97", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"\u521b\u5efaTESS NG", None))
        self.btnShowXodr.setText(QCoreApplication.translate("TESS_API_EXAMPLEClass", u"开始创建TESS NG路网", None))

    def change_progress(self, pb, value, network_info=None, error=False):
        if error:
            # 导入失败，窗体不显示，错误提示
            self.pb.setVisible(False)
            self.text_label_1.setVisible(False)
            self.groupBox_2.setVisible(False)
            self.btnOpenNet.setEnabled(True)
            QMessageBox.warning(None, "提示信息", "路网解析错误，请联系开发者")
            return
        pb.setValue(value)
        if not network_info:
            self.pb.setVisible(True)
        else:
            self.pb.setVisible(False)
            # 导入完成后，部分窗体展示
            self.text_label_1.setVisible(True)
            # 提取 network_info 并显示
            context = "\n".join(
                [
                    f"road_id: {road_id}, lanes: {[section['all'] for _, section in sorted(network_info['roads_info'][road_id]['lane_sections'].items(), key=lambda b: b[0])]}"
                    for road_id, road_info in network_info["roads_info"].items()
                ]
            )
            self.txtMessage1.setText(context)
            self.groupBox_2.setVisible(True)
            self.btnOpenNet.setEnabled(True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytessng",
    version="0.1.9",
    author="yang",
    author_email="17315487709@163.com",
    description="tessng with python3.6",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    install_requires=[
        'opendrive2tessng==0.1.6',
        'tessng2other',
        'PySide2==5.15.2.1',
    ],
    python_requires='>=3.6, <3.8',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={  # Optional
            'pytessng': ['*.dll', '*.pyd', '*.py', '*.pyi'],
        },
)



# -*- coding: utf-8 -*-

from PySide2.QtWidgets import QDockWidget
from pytessng.Tessng import TessPlugin
from TESS_API_EXAMPLE import *
from MyNet import MyNet


# 用户插件，继承自TessPlugin
class MyPlugin(TessPlugin):
    def __init__(self):
        super(MyPlugin, self).__init__()
        self.mNetInf = None
        self.mSimuInf = None
        self.dockWidget = None

    def initGui(self):
        # 在TESS NG主界面上增加 QDockWidget对象
        self.examleWindow = TESS_API_EXAMPLE()

        iface = tngIFace()
        win = iface.guiInterface().mainWindow()

        dockWidget = QDockWidget("opendrive文件导入", win)
        dockWidget.setObjectName("mainDockWidget")
        dockWidget.setFeatures(QDockWidget.NoDockWidgetFeatures)
        dockWidget.setAllowedAreas(Qt.LeftDockWidgetArea)
        dockWidget.setWidget(self.examleWindow.centralWidget())
        self.dockWidget = dockWidget
        iface.guiInterface().addDockWidgetToMainWindow(Qt.DockWidgetArea(1), dockWidget)

    # 过载父类方法，在 TESS NG工厂类创建TESS NG对象时调用
    def init(self):
        self.initGui()
        # self.mNetInf = PyCustomerNet()
        self.mNetInf = MyNet()

    # 过载父类方法，返回插件路网子接口，此方法由TESS NG调用
    def customerNet(self):
        return self.mNetInf

    # 过载父类方法，返回插件仿真子接口，此方法由TESS NG调用
    def customerSimulator(self):
        return self.mSimuInf
